/*
 *  =======================================================================
 *  Compilation:  javac Matrix.java
 *  Execution:    java Matrix
 *
 *  Basic implementation of an M-by-N matrix and matrix operations
 *
 *  Written by: Mark Austin                                  September 2009
 *  =======================================================================
 */

package objects;

import java.lang.Math;
import java.util.*;
import java.io.*;
import java.text.*;

public class Matrix {
    private String   sName;  // name of array ...
    private int    iNoRows;  // # rows
    private int iNoColumns;  // # columns
    private double[][] data; // the 2d array

    // Constructors for M-by-N matrices of 0's

    public Matrix(int iNoRows, int iNoColumns ) {
        this.iNoRows    = iNoRows;
        this.iNoColumns = iNoColumns;
        this.data       = new double[ iNoRows ][ iNoColumns ];
    }

    public Matrix( String sName, int iNoRows, int iNoColumns ) {
        this.sName      = sName;
        this.iNoRows    = iNoRows;
        this.iNoColumns = iNoColumns;
        this.data       = new double[ iNoRows ][ iNoColumns ];
    }

    // ======================================
    // Set/get name for data array
    // ======================================

    public void setName ( String sName ) {
       this.sName = sName;
    }

    public String getName() {
       return this.sName;
    }

    // ======================================
    // Retrieve and set matrix element values
    // ======================================

    public double getElement( int i, int j ) {
        double returnValue;

        if( i < 0 || j < 0 || i >= iNoRows || j >= iNoColumns ) throw new
            RuntimeException("*** In getElement(): Array element index out of range");

        returnValue = data[i][j];
        return returnValue;
    }

    public void setElement( int i, int j, double value )  {

        if( i < 0 || j < 0 || i >= iNoRows || j >= iNoColumns ) throw new
            RuntimeException("*** In setElement(): Array element index out of range");

        this.data[i][j] = value;
    }

    // Create matrices based on 2d arrays

    public Matrix(String sName, double[][] data) {
        this.sName      = sName;
        this.iNoRows    = data.length;
        this.iNoColumns = data[0].length;
        this.data       = new double[ iNoRows ][ iNoColumns ];
        for (int i = 0; i < iNoRows; i++)
            for (int j = 0; j < iNoColumns; j++)
                    this.data[i][j] = data[i][j];
    }

    public Matrix(double[][] data) {
        this.iNoRows    = data.length;
        this.iNoColumns = data[0].length;
        this.data       = new double[ iNoRows ][ iNoColumns ];
        for (int i = 0; i < iNoRows; i++)
            for (int j = 0; j < iNoColumns; j++)
                    this.data[i][j] = data[i][j];
    }

    // copy constructor

    public Matrix( Matrix A ) {
        this(A.data);
    }

    // create and return a random M-by-N matrix with values between 0 and 1

    public static Matrix random(int iNoRows, int iNoColumns ) {
        Matrix A = new Matrix( iNoRows, iNoColumns );
        for (int i = 0; i < iNoRows; i++)
            for (int j = 0; j < iNoColumns; j++)
                A.data[i][j] = Math.random();
        return A;
    }

    // create and return the N-by-N identity matrix

    public static Matrix identity(int iNoRows ) {

        Matrix I = new Matrix( iNoRows, iNoRows);
        for (int i = 0; i < iNoRows; i++)
            I.data[i][i] = 1;
        return I;
    }

    // swap rows i and j

    public void swap(int i, int j) {
        double[] temp = data[i];
        data[i]       = data[j];
        data[j]       = temp;
    }

    // create and return the transpose of the invoking matrix

    public Matrix transpose() {
        Matrix A = new Matrix( iNoColumns, iNoRows );
        for (int i = 0; i < iNoRows; i++)
            for (int j = 0; j < iNoColumns; j++)
                A.data[j][i] = this.data[i][j];
        return A;
    }

    // return C = A + B

    public Matrix plus(Matrix B) {
        Matrix A = this;

        // Check compatibility of matrix dimensions ....

        if (B.iNoRows != A.iNoRows || B.iNoColumns != A.iNoColumns ) throw
            new RuntimeException("Illegal matrix dimensions.");

        // Create new matrix 

        Matrix C = new Matrix( iNoRows, iNoColumns );

        for (int i = 0; i < iNoRows; i++)
            for (int j = 0; j < iNoColumns; j++)
                C.data[i][j] = A.data[i][j] + B.data[i][j];

        return C;
    }

    // return C = A - B

    public Matrix minus(Matrix B) {
        Matrix A = this;

        // Check compatibility of matrix dimensions ....

        if (B.iNoRows != A.iNoRows || B.iNoColumns != A.iNoColumns ) throw
            new RuntimeException("Illegal matrix dimensions.");

        Matrix C = new Matrix( iNoRows, iNoColumns );

        for (int i = 0; i < iNoRows; i++)
            for (int j = 0; j < iNoColumns; j++)
                C.data[i][j] = A.data[i][j] - B.data[i][j];
        return C;
    }

    // Test: does A = B exactly?

    public boolean equals(Matrix B) {
        Matrix A = this;

        if (B.iNoRows != A.iNoRows || B.iNoColumns != A.iNoColumns ) throw
            new RuntimeException("Illegal matrix dimensions.");

        for (int i = 0; i < iNoRows; i++)
            for (int j = 0; j < iNoColumns; j++)
                if (A.data[i][j] != B.data[i][j]) return false;

        return true;
    }

    // return C = A * B

    public Matrix times(Matrix B) {
        Matrix A = this;

        if ( A.iNoColumns != B.iNoRows ) throw
             new RuntimeException("Illegal matrix dimensions.");

        Matrix C = new Matrix( A.iNoRows, B.iNoColumns );

        for (int i = 0; i < C.iNoRows; i++)
            for (int j = 0; j < C.iNoColumns; j++)
                for (int k = 0; k < A.iNoColumns; k++)
                    C.data[i][j] += (A.data[i][k] * B.data[k][j]);
        return C;
    }

    // =========================================================
    // return x = A^-1 b, assuming A is square and has full rank
    // =========================================================

    public Matrix solve(Matrix rhs) {

        if (iNoRows != iNoColumns || rhs.iNoRows != iNoColumns || rhs.iNoColumns != 1)
            throw new RuntimeException("Illegal matrix dimensions.");

        // create copies of the data

        Matrix A = new Matrix(this);
        Matrix b = new Matrix(rhs);

        // Gaussian elimination with partial pivoting

        for (int i = 0; i < iNoColumns; i++) {

            // find pivot row and swap

            int max = i;
            for (int j = i + 1; j < iNoColumns; j++)
                if (Math.abs(A.data[j][i]) > Math.abs(A.data[max][i]))
                    max = j;
            A.swap(i, max);
            b.swap(i, max);

            // singular

            if (A.data[i][i] == 0.0) throw
                new RuntimeException("Matrix is singular.");

            // pivot within b

            for (int j = i + 1; j < iNoColumns; j++)
                b.data[j][0] -= b.data[i][0] * A.data[j][i] / A.data[i][i];

            // pivot within A

            for (int j = i + 1; j < iNoColumns; j++) {
                double m = A.data[j][i] / A.data[i][i];
                for (int k = i+1; k < iNoColumns; k++) {
                    A.data[j][k] -= A.data[i][k] * m;
                }
                A.data[j][i] = 0.0;
            }
        }

        // back substitution

        Matrix x = new Matrix(iNoColumns, 1);
        for (int j = iNoColumns - 1; j >= 0; j--) {
            double t = 0.0;
            for (int k = j + 1; k < iNoColumns; k++)
                t += A.data[j][k] * x.data[k][0];
            x.data[j][0] = (b.data[j][0] - t) / A.data[j][j];
        }

        return x;
    }

    // ============================================
    // print matrix to standard output
    // ============================================

    // Define NoColumns to be a "constant" .... 

    public final static int NoColumns = 6;

    public void print() {
       int ii, ij, ib;   
       int iNoBlocks;
       int iFirstColumn, iLastColumn;

       // Compute no of blocks of rows to be printed .....

       if( this.iNoColumns % NoColumns == 0 )
          iNoBlocks = this.iNoColumns/NoColumns;
       else
          iNoBlocks = this.iNoColumns/NoColumns + 1;

       // Loop over the number of blocks ...

       for( ib = 1; ib <= iNoBlocks; ib++ ) { 
          iFirstColumn = (ib-1)*NoColumns + 1;
          iLastColumn  = Math.min ( ib*NoColumns, this.iNoColumns );

          // Print title of matrix at top of each block ....
       
          if( this.sName != null ) 
              System.out.println("\nMatrix: " + this.sName );
          else 
              System.out.println("\nMatrix: Untitled");

          // Label row and column nos */

          System.out.printf ("row/col          ");
          for( ii = iFirstColumn; ii <= iLastColumn; ii++ )
              System.out.printf("       %3d   ", ii);
          System.out.printf("\n");

          // Print contents of matrix  ....

          for( ii = 1; ii <= iNoRows; ii++ ) {
              System.out.printf(" %3d ", ii);
              System.out.printf("         ");
              for( ij  = iFirstColumn; ij <= iLastColumn; ij++)
                  System.out.printf(" %12.5e", this.data[ ii-1 ][ ij-1 ] );
              System.out.printf("\n");
          }
       }
    }

    // ============================================
    // Convert array to string format ....
    // ============================================

    public String toString() {
        String s = "Array: " + this.sName + "\n";

        for (int i = 0; i < iNoRows; i++) {
            for (int j = 0; j < iNoColumns; j++) {
                s += " " + data[i][j];
                if (j == iNoColumns-1 )
                    s += "\n";
            }

            System.out.println();
        }

        return s;
    }

    // ================================================
    // Exercise methods in the Matrix class ....
    // ================================================

    public static void main(String[] args) {

        double[][] d = { { 1, 2, 3 }, { 4, 5, 6 }, { 9, 1, 3} };
        Matrix D = new Matrix(d);

        D.print();        
        System.out.println();

        Matrix A = Matrix.random(5, 5);
        A.print(); 
        System.out.println();

        A.swap(1, 2);
        A.print(); 
        System.out.println();

        Matrix B = A.transpose();
        B.print(); 
        System.out.println();

        Matrix C = Matrix.identity(5);
        C.print(); 
        System.out.println();

        A.plus(B).print();
        System.out.println();

        B.times(A).print();
        System.out.println();

        // shouldn't be equal since AB != BA in general    

        System.out.println(A.times(B).equals(B.times(A)));
        System.out.println();

        // Create a large identity martix

        Matrix large = Matrix.identity(10);
        large.setName("Identity!"); 
        large.print(); 

        // Solve a simple set of equations 

        Matrix a = new Matrix ( "A", 3, 3 );
        a.setElement( 0, 0,  1 );
        a.setElement( 0, 1,  2 );
        a.setElement( 0, 2,  3 );

        a.setElement( 1, 0,  5 );
        a.setElement( 1, 1,  6 );
        a.setElement( 1, 2,  7 );

        a.setElement( 2, 0,  2 );
        a.setElement( 2, 1,  4 );
        a.setElement( 2, 2,  8 );
        a.print();

        Matrix b = new Matrix ( "B", 3, 1 );
        b.setElement( 0, 0,  6 );
        b.setElement( 1, 0, 18 );
        b.setElement( 2, 0, 14 );
        b.print();

        // Compute solution to equations ....

        Matrix x = a.solve( b );
        x.setName("x");

        // Print and check result ...

        x.print();
        a.times(x).print();
    }
}
