/*
 * =========================================================================
 * Spreadsheet.java: An implementation of Spreadsheet as described in spec 
 *
 * Example: a1=1;b1=c1+a1;c1=b1-f1;d1=b1-2;e1=c1;f1=-1;a2=b2;b2=c2;c2=d2;
 *          d2=e2;e2=a2;a3=d3;b3=a3+c3;d3=b3;c3=e3;
 * =========================================================================
 */

package spreadsheet;

import spreadsheet.parser.*;
import java.util.*;
import java.util.concurrent.ConcurrentLinkedQueue;

public class Spreadsheet implements SpreadsheetInterface {

   private int numRows, numColumns; // no. of rows and cols for spreadsheet
   private Map<String, Cell> cells; // collection of all cells in spreadsheet
   private String lastCellLocation; // stores location of last cell accessed

   /**
     * dirtyCells: list contains cells marked dirty ie. value is outdated and
     * needs to be recalculated - returns expression in cell if dirty, computes
     * value otherwise; keeps track of visited cells for loop checking algorithm
     */

   private Queue<String> dirtyCells;

   // Default class constructor (for GUI)

   public Spreadsheet(int numRows, int numColumns) {
      this.numRows    = numRows;           // update no of rows
      this.numColumns = numColumns;        // update no of cols
      cells = new HashMap<String, Cell>(); // init spreadsheet of cells
   }

   // Returns the number of rows in the spreadsheet

   public int getRowCount() {
      return numRows;
   }

   // Returns the number of columns in the spreadsheet

   public int getColumnCount() {
      return numColumns;
   }

   // Removes instances of cellLocation as parent in the children cells

   private void removeOldParents(String cellLocation) {
      for (String parent : cells.get(cellLocation).getParent()) {
           cells.get(parent).removeChildren(cellLocation);
      }
   }

   /**
     * returns the expression stored at the cell at location pre: location is a
     * lowercase String
     */

	public String getExpression(String location) {
		// so that a1 / A1 will both refer to the same cell a1
		location = location.toLowerCase();
		/**
		 * check if location = valid cell reference && referenced cell is
		 * non-empty
		 */
		return Expression.isCellLocation(location)
				&& cells.containsKey(location) ? cells.get(location)
				.getExpression() : null;
	}

	// returns the value associated with the computed stored expression

	public Object getValue(String location) {
		// so that a1 / A1 will both refer to the same cell a1
		location = location.toLowerCase();
		/**
		 * check if location = valid cell reference && referenced cell is
		 * non-empty
		 */
		return Expression.isCellLocation(location)
				&& cells.containsKey(location) ? cells.get(location).getValue()
				: null;
	}

	/**
	 * sets the expression of the cell at location and attempts to compute its
	 * current value. Any parent cell that has either acquired the cell as a
	 * child or no longer has the cell as a child should be modified
	 * accordingly.
	 */

	public void setExpression(String location, String expression) {
		// so that a1 / A1 will both reference the same cell a1
		location = location.toLowerCase();

		// check if location = valid cell reference
		if (Expression.isCellLocation(location)) {
			// update last cell location
			lastCellLocation = location;

			// add new cell to cells if current cell is undefined
			if (!cells.containsKey(location)) {
				// add new Cell() to Map cells
				cells.put(location, new Cell());
			}

			// get current cell @ location
			Cell currentCell = cells.get(location);
			/**
			 * // if expression is unchanged, exit => save on unnecessary
			 * recalculation if (currentCell.getExpression().equals(expression))
			 * { return; }
			 */
			// update expression of cell
			currentCell.setExpression(expression);

			/**
			 * -----------------------------------------------------------
			 * update parent-child relationship for currentCell
			 * -----------------------------------------------------------
			 */
			// remove itself as a child of its previous parent(s)
			removeOldParents(location);

			// set { referenced cells } as its new parent(s)
			currentCell.setParent(Expression.getReferences(expression));

			// set itself as the child of its referenced cells
			for (String nextCellLocation : currentCell.getParent()) {
				// add new empty cell if it does not exist in cells;
				if (!cells.containsKey(nextCellLocation)) {
					cells.put(nextCellLocation, new Cell());
				}

				// add current cell to nextCell's parent
				cells.get(nextCellLocation).addChildren(location);
			}
			// ___________________________________________________________

			// reset dirty flag on children
			for (String nextCellLocation : currentCell.getChildren()) {
				cells.get(nextCellLocation).setDirty(false);
			}

			recompute();
		}
	}

	/**
	 * computes the value of all cells whose expressions were changed since the
	 * last call to this method. This also requires the computation of the
	 * values of any dependent cells. Cells forming a loop are assigned the
	 * value Spreadsheet.LOOP.
	 */

	public void recompute() {
		// recompute last active changed cell

		recomputeValue();

		// check if location = valid cell reference && lastCell exists in cells

		if (Expression.isCellLocation(lastCellLocation)
				&& cells.containsKey(lastCellLocation)) {
			// get last cell
			Cell lastCell = cells.get(lastCellLocation);

			// create a new empty queue; avoids ConcurrentModificationException
			Queue<String> queue = new ConcurrentLinkedQueue<String>();

			// add lastCell's children to queue
			queue.addAll(lastCell.getChildren());

			// process elements in the queue
			for (String queueElement : queue) {
				// set cell's expression if it is not dirty
				if (!cells.get(queueElement).isDirty()) {
					setExpression(queueElement, cells.get(queueElement)
							.getExpression());
				}

				// remove queueElement from the front of the queue
				queue.remove();
			}
		}
	}

	// recomputes value for lastCell only
	private void recomputeValue() {
		// get last cell; assumes lastCellLocation is valid
		Cell lastCell = cells.get(lastCellLocation);

		// check if the expression in lastCell is a valid formula
		if (Expression.isFormula(lastCell.getExpression())) {
			// construct map { referenced parent cells -> value }
			Map<String, Double> lookUpCellValue = new HashMap<String, Double>();
			for (String parent : lastCell.getParent()) {
				if (cells.containsKey(parent)
						&& cells.get(parent).getValue() instanceof Double) {
					lookUpCellValue.put(parent, (Double) cells.get(parent)
							.getValue());
				} else if (!cells.containsKey(parent)) {
					// otherwise add new Cell() to Map cells
					cells.put(parent, new Cell());
				}
			}

			dirtyCells = new ConcurrentLinkedQueue<String>();

			// detect circular references
			checkLOOP(lastCellLocation);

			/**
			 * returns expression if cell is not marked as 'dirty'; leaves it as
			 * #LOOP otherwise
			 */
			if (!lastCell.isDirty()) {
				lastCell.setValue(Expression.computeValue(lastCell
						.getExpression(), lookUpCellValue));
			}
		} else {
			// lastCell does not have a valid formula -> value = expr
			lastCell.setValue(lastCell.getExpression());
		}
	}

   // Implements loop checking via DFS recursion

   private void checkLOOP(String cellLocation) {

      // Add cellLocation to list of dirty cells

         dirtyCells.add(cellLocation);

      // Default: return computed value for cell

	 cells.get(cellLocation).setDirty(false);

      // Check if children cells form a cycle

         for (String child : cells.get(cellLocation).getChildren()) {
			if (dirtyCells.contains(child)) {
				// refresh dirtyCells to search for cycles in next branch
				Iterator<String> it = dirtyCells.iterator();
				String nextDirtyCell;
				while (it.hasNext() && (nextDirtyCell = it.next()) != child) {
					dirtyCells.remove();
				}

				// set cellValues to #LOOP
				for (String dirtyCellLocation : dirtyCells) {
					Cell dirtyCell = cells.get(dirtyCellLocation);
					dirtyCell.setDirty(true);
					dirtyCell.setValue(LOOP);
				}

				return;
			}

			// Recursive call to check loops on next child

			checkLOOP(child);
			dirtyCells.remove(child);
         }
    }
}
